import logging
from collections.abc import Iterable
from dataclasses import dataclass

from databricks.labs.lsql.backends import SqlBackend

from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.hive_metastore import TablesCrawler

logger = logging.getLogger(__name__)


@dataclass
class TableSize:
    catalog: str
    database: str
    name: str
    size_in_bytes: int


class TableSizeCrawler(CrawlerBase):
    def __init__(self, backend: SqlBackend, schema, include_databases: list[str] | None = None):
        """
        Initializes a TablesSizeCrawler instance.

        Args:
            backend (SqlBackend): The SQL Execution Backend abstraction (either REST API or Spark)
            schema: The schema name for the inventory persistence.
        """
        # pylint: disable-next=import-error,import-outside-toplevel
        from pyspark.sql.session import SparkSession  # type: ignore[import-not-found]

        self._backend = backend
        super().__init__(backend, "hive_metastore", schema, "table_size", TableSize)
        self._tables_crawler = TablesCrawler(backend, schema, include_databases)
        self._spark = SparkSession.builder.getOrCreate()

    def _crawl(self) -> Iterable[TableSize]:
        """Crawls and lists tables using table crawler
        Identifies DBFS root tables and calculates the size for these.
        """
        for table in self._tables_crawler.snapshot():
            if not table.kind == "TABLE":
                continue
            if not table.is_dbfs_root:
                continue
            size_in_bytes = self._safe_get_table_size(table.key)
            if size_in_bytes is None:
                continue  # table does not exist anymore or is corrupted

            yield TableSize(
                catalog=table.catalog, database=table.database, name=table.name, size_in_bytes=size_in_bytes
            )

    def _try_fetch(self) -> Iterable[TableSize]:
        """Tries to load table information from the database or throws TABLE_OR_VIEW_NOT_FOUND error"""
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield TableSize(*row)

    def _safe_get_table_size(self, table_full_name: str) -> int | None:
        logger.debug(f"Evaluating {table_full_name} table size.")
        try:
            # refresh table statistics to avoid stale stats in HMS
            self._backend.execute(f"ANALYZE table {escape_sql_identifier(table_full_name)} compute STATISTICS NOSCAN")
            # pylint: disable-next=protected-access
            return self._spark._jsparkSession.table(table_full_name).queryExecution().analyzed().stats().sizeInBytes()
        except Exception as e:  # pylint: disable=broad-exception-caught
            if "[TABLE_OR_VIEW_NOT_FOUND]" in str(e) or "[DELTA_TABLE_NOT_FOUND]" in str(e):
                logger.warning(f"Failed to evaluate {table_full_name} table size. Table not found.")
                return None
            if "[DELTA_INVALID_FORMAT]" in str(e):
                logger.warning(
                    f"Unable to read Delta table {table_full_name}, please check table structure and try again."
                )
                return None
            if "[DELTA_MISSING_TRANSACTION_LOG]" in str(e):
                logger.warning(f"Delta table {table_full_name} is corrupted: missing transaction log.")
                return None
            logger.error(f"Failed to evaluate {table_full_name} table size: ", exc_info=True)

        return None



import json
import logging

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.installer import InstallState
from databricks.labs.blueprint.tui import Prompts
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound
from databricks.sdk.service import compute
from databricks.sdk.service.sql import GetWorkspaceWarehouseConfigResponse


logger = logging.getLogger(__name__)

# Known prefixes for external HMS spark attributes
ext_hms_prefixes = [
    "spark_conf.spark.sql.hive.metastore",
    "spark_conf.spark.hadoop.hive.metastore",
    "spark_conf.spark.hadoop.javax.jdo.option",
    "spark_conf.spark.databricks.hive.metastore",
]


class ClusterPolicyInstaller:
    def __init__(self, installation: Installation, ws: WorkspaceClient, prompts: Prompts):
        self._ws = ws
        self._installation = installation
        self._prompts = prompts

    @staticmethod
    def _policy_config(value: str):
        return {"type": "fixed", "value": value}

    def has_ext_hms(self) -> bool:
        policies_with_external_hms = list(self._get_cluster_policies_with_external_hive_metastores())
        if len(policies_with_external_hms) > 0:
            return True
        warehouse_config = self._get_warehouse_config_with_external_hive_metastore()
        if warehouse_config is not None:
            return True
        return False

    def create(self, inventory_database: str) -> tuple[str, str, dict, str | None]:
        instance_profile = ""
        spark_conf_dict = {}
        # get instance pool id to be put into the cluster policy
        instance_pool_id = self._get_instance_pool_id()
        policies_with_external_hms = list(self._get_cluster_policies_with_external_hive_metastores())
        if len(policies_with_external_hms) > 0 and self._prompts.confirm(
            "We have identified one or more cluster policies set up for an external metastore. "
            "Would you like to set UCX to connect to the external metastore?"
        ):
            logger.info("Setting up an external metastore")
            cluster_policies = {conf.name: conf.definition for conf in policies_with_external_hms}
            if len(cluster_policies) >= 1:
                cluster_policy = json.loads(self._prompts.choice_from_dict("Choose a cluster policy", cluster_policies))
                instance_profile, spark_conf_dict = self._extract_external_hive_metastore_conf(cluster_policy)
        else:
            warehouse_config = self._get_warehouse_config_with_external_hive_metastore()
            if warehouse_config and self._prompts.confirm(
                "We have identified the workspace warehouse is set up for an external metastore. "
                "Would you like to set UCX to connect to the external metastore?"
            ):
                logger.info("Setting up an external metastore")
                instance_profile, spark_conf_dict = self._extract_external_hive_metastore_sql_conf(warehouse_config)
        policy_name = f"Unity Catalog Migration ({inventory_database}) ({self._ws.current_user.me().user_name})"
        policies = self._ws.cluster_policies.list()
        for policy in policies:
            if policy.name == policy_name:
                logger.info(f"Cluster policy {policy_name} already present, reusing the same.")
                policy_id = policy.policy_id
                assert policy_id is not None
                return policy_id, instance_profile, spark_conf_dict, instance_pool_id
        logger.info("Creating UCX cluster policy.")
        policy_id = self._ws.cluster_policies.create(
            name=policy_name,
            definition=self._definition(spark_conf_dict, instance_profile, instance_pool_id),
            description="Custom cluster policy for Unity Catalog Migration (UCX)",
        ).policy_id
        assert policy_id is not None
        return (
            policy_id,
            instance_profile,
            spark_conf_dict,
            instance_pool_id,
        )

    def _get_instance_pool_id(self) -> str | None:
        try:
            instance_pool_id = self._prompts.question(
                "Instance pool id to be set in cluster policy for all workflow clusters", default="None"
            )
        except OSError:
            # when unit test v0.15.0_added_cluster_policy.py MockPrompts cannot be injected to ClusterPolicyInstaller
            # return None to pass the test
            return None
        if instance_pool_id.lower() == "none":
            return None
        try:
            self._ws.instance_pools.get(instance_pool_id)
            return instance_pool_id
        except NotFound:
            logger.warning(
                f"Instance pool id {instance_pool_id} does not exist. Will not set instance pool in the cluster policy. You can manually edit the cluster policy after installation."
            )
            return None

    def _definition(self, conf: dict, instance_profile: str | None, instance_pool_id: str | None) -> str:
        latest_lts_dbr = self._ws.clusters.select_spark_version(latest=True)
        node_type_id = self._ws.clusters.select_node_type(local_disk=True, min_memory_gb=32)
        policy_definition = {
            "spark_version": self._policy_config(latest_lts_dbr),
            "node_type_id": self._policy_config(node_type_id),
        }
        for key, value in conf.items():
            policy_definition[f"spark_conf.{key}"] = self._policy_config(value)
        # set the availability to on demand
        if self._ws.config.is_aws:
            if instance_profile:
                policy_definition["aws_attributes.instance_profile_arn"] = self._policy_config(instance_profile)
            policy_definition["aws_attributes.availability"] = self._policy_config(
                compute.AwsAvailability.ON_DEMAND.value
            )
            policy_definition["aws_attributes.zone_id"] = self._policy_config("auto")
        elif self._ws.config.is_azure:
            policy_definition["azure_attributes.availability"] = self._policy_config(
                compute.AzureAvailability.ON_DEMAND_AZURE.value
            )
        else:
            policy_definition["gcp_attributes.availability"] = self._policy_config(
                compute.GcpAvailability.ON_DEMAND_GCP.value
            )
        if instance_pool_id:
            policy_definition["instance_pool_id"] = self._policy_config(instance_pool_id)
            # 'node_type_id' cannot be supplied when an instance pool ID is provided
            policy_definition.pop("node_type_id")
            # 'availability' cannot be supplied when an instance pool ID is provided
            policy_definition.pop("aws_attributes.availability", "")
            # 'zone_id' cannot be supplied when an instance pool ID is provided
            policy_definition.pop("aws_attributes.zone_id", "")
            policy_definition.pop("azure_attributes.availability", "")
            policy_definition.pop("gcp_attributes.availability", "")
        return json.dumps(policy_definition)

    @staticmethod
    def _extract_external_hive_metastore_conf(cluster_policy):
        spark_conf_dict = {}
        instance_profile = None
        if cluster_policy.get("aws_attributes.instance_profile_arn") is not None:
            instance_profile = cluster_policy.get("aws_attributes.instance_profile_arn").get("value")
            logger.info(f"Instance Profile is Set to {instance_profile}")
        for key in cluster_policy.keys():
            for known_prefix in ext_hms_prefixes:
                if key.startswith(known_prefix):
                    spark_conf_dict[key[11:]] = cluster_policy[key]["value"]
        return instance_profile, spark_conf_dict

    def _get_cluster_policies_with_external_hive_metastores(self):
        for policy in self._ws.cluster_policies.list():
            def_json = json.loads(policy.definition)
            glue_node = def_json.get("spark_conf.spark.databricks.hive.metastore.glueCatalog.enabled")
            if glue_node is not None and glue_node.get("value") == "true":
                yield policy
                continue
            for key in def_json.keys():
                if key.startswith("spark_conf.spark.sql.hive.metastore"):
                    yield policy
                    break

    @staticmethod
    def _extract_external_hive_metastore_sql_conf(sql_config: GetWorkspaceWarehouseConfigResponse):
        spark_conf_dict: dict[str, str] = {}
        instance_profile = None
        if sql_config.instance_profile_arn is not None:
            instance_profile = sql_config.instance_profile_arn
            logger.info(f"Instance Profile is Set to {instance_profile}")
        if sql_config.data_access_config is None:
            return instance_profile, spark_conf_dict
        for conf in sql_config.data_access_config:
            if conf.key is None:
                continue
            if conf.value is None:
                continue
            if (
                conf.key.startswith("spark.sql.hive.metastore")
                or conf.key.startswith("spark.hadoop.javax.jdo.option")
                or conf.key.startswith("spark.databricks.hive.metastore")
                or conf.key.startswith("spark.hadoop.hive.metastore.glue")
            ):
                spark_conf_dict[conf.key] = conf.value
        return instance_profile, spark_conf_dict

    def _get_warehouse_config_with_external_hive_metastore(self) -> GetWorkspaceWarehouseConfigResponse | None:
        sql_config = self._ws.warehouses.get_workspace_warehouse_config()
        if sql_config.data_access_config is None:
            return None
        for conf in sql_config.data_access_config:
            if conf.key is None:
                continue
            is_glue = conf.key.startswith("spark.databricks.hive.metastore.glueCatalog.enabled")
            if conf.key.startswith("spark.sql.hive.metastore") or is_glue:
                return sql_config
        return None

    def update_job_policy(self, state: InstallState, policy_id: str):
        if not state.jobs:
            logger.error("No jobs found in states")
            return
        for _, job_id in state.jobs.items():
            try:
                job = self._ws.jobs.get(job_id)
                job_settings = job.settings
                assert job.job_id is not None
                assert job_settings is not None
                if job_settings.job_clusters is None:
                    # if job_clusters is None, it means override cluster is being set
                    # and hence policy should not be applied
                    return
            except NotFound:
                logger.error(f"Job id {job_id} not found. Please check if the job is present in the workspace")
                continue
            try:
                job_clusters = []
                for cluster in job_settings.job_clusters:
                    assert cluster.new_cluster is not None
                    cluster.new_cluster.policy_id = policy_id
                    job_clusters.append(cluster)
                job_settings.job_clusters = job_clusters
                self._ws.jobs.update(job.job_id, new_settings=job_settings)
            except NotFound:
                logger.error(f"Job id {job_id} not found.")
                continue



import logging

from databricks.labs.ucx.contexts.workflow_task import RuntimeContext
from databricks.labs.ucx.framework.tasks import Workflow, job_task


logger = logging.getLogger(__name__)


class Assessment(Workflow):
    def __init__(self):
        super().__init__('assessment')

    @job_task(notebook="hive_metastore/tables.scala")
    def crawl_tables(self, ctx: RuntimeContext):
        """Iterates over all tables in the Hive Metastore of the current workspace and persists their metadata, such
        as _database name_, _table name_, _table type_, _table location_, etc., in the Delta table named
        `$inventory_database.tables`. Note that the `inventory_database` is set in the configuration file. The metadata
        stored is then used in the subsequent tasks and workflows to, for example,  find all Hive Metastore tables that
        cannot easily be migrated to Unity Catalog."""

    @job_task
    def crawl_udfs(self, ctx: RuntimeContext):
        """Iterates over all UDFs in the Hive Metastore of the current workspace and persists their metadata in the
        table named `$inventory_database.udfs`. This inventory is currently used when scanning securable objects for
        issues with grants that cannot be migrated to Unit Catalog."""
        ctx.udfs_crawler.snapshot()

    @job_task(job_cluster="tacl")
    def setup_tacl(self, ctx: RuntimeContext):
        """(Optimization) Starts `tacl` job cluster in parallel to crawling tables."""

    @job_task(depends_on=[crawl_tables, crawl_udfs, setup_tacl], job_cluster="tacl")
    def crawl_grants(self, ctx: RuntimeContext):
        """Scans all securable objects for permissions that have been assigned: this include database-level permissions,
        as well permissions directly configured on objects in the (already gathered) table and UDF inventories. The
        captured information is stored in the `$inventory_database.grants` inventory table for further use during the
        migration of legacy ACLs to Unity Catalog permissions.

        Note: This job runs on a separate cluster (named `tacl`) as it requires the proper configuration to have the Table
        ACLs enabled and available for retrieval."""
        ctx.grants_crawler.snapshot()

    @job_task(depends_on=[crawl_tables])
    def estimate_table_size_for_migration(self, ctx: RuntimeContext):
        """Scans the previously created Delta table named `$inventory_database.tables` and locate tables that cannot be
        "synced". These tables will have to be cloned in the migration process.
        Assesses the size of these tables and create `$inventory_database.table_size` table to list these sizes.
        The table size is a factor in deciding whether to clone these tables."""
        ctx.table_size_crawler.snapshot()

    @job_task
    def crawl_mounts(self, ctx: RuntimeContext):
        """Defines the scope of the _mount points_ intended for migration into Unity Catalog. As these objects are not
        compatible with the Unity Catalog paradigm, a key component of the migration process involves transferring them
        to Unity Catalog External Locations.

        The assessment involves scanning the workspace to compile a list of all existing mount points and subsequently
        storing this information in the `$inventory.mounts` table. This is crucial for planning the migration."""
        ctx.mounts_crawler.snapshot()

    @job_task(depends_on=[crawl_mounts, crawl_tables])
    def guess_external_locations(self, ctx: RuntimeContext):
        """Determines the shared path prefixes of all the tables. Specifically, the focus is on identifying locations that
        utilize mount points. The goal is to identify the _external locations_ necessary for a successful migration and
        store this information in the `$inventory.external_locations` table.

        The approach taken in this assessment involves the following steps:
          - Extracting all the locations associated with tables that do not use DBFS directly, but a mount point instead
          - Scanning all these locations to identify folders that can act as shared path prefixes
          - These identified external locations will be created subsequently prior to the actual table migration"""
        ctx.external_locations.snapshot()

    @job_task
    def assess_jobs(self, ctx: RuntimeContext):
        """Scans through all the jobs and identifies those that are not compatible with UC. The list of all the jobs is
        stored in the `$inventory.jobs` table.

        It looks for:
          - Clusters with Databricks Runtime (DBR) version earlier than 11.3
          - Clusters using Passthrough Authentication
          - Clusters with incompatible Spark config tags
          - Clusters referencing DBFS locations in one or more config options
        """
        ctx.jobs_crawler.snapshot()

    @job_task
    def assess_clusters(self, ctx: RuntimeContext):
        """Scan through all the clusters and identifies those that are not compatible with UC. The list of all the clusters
        is stored in the`$inventory.clusters` table.

        It looks for:
          - Clusters with Databricks Runtime (DBR) version earlier than 11.3
          - Clusters using Passthrough Authentication
          - Clusters with incompatible spark config tags
          - Clusters referencing DBFS locations in one or more config options
        """
        ctx.clusters_crawler.snapshot()

    @job_task
    def assess_pipelines(self, ctx: RuntimeContext):
        """This module scans through all the Pipelines and identifies those pipelines which has Azure Service Principals
        embedded (who has been given access to the Azure storage accounts via spark configurations) in the pipeline
        configurations.

        It looks for:
          - all the pipelines which has Azure Service Principal embedded in the pipeline configuration

        Subsequently, a list of all the pipelines with matching configurations are stored in the
        `$inventory.pipelines` table."""
        ctx.pipelines_crawler.snapshot()

    @job_task
    def assess_incompatible_submit_runs(self, ctx: RuntimeContext):
        """This module scans through all the Submit Runs and identifies those runs which may become incompatible after
        the workspace attachment.

        It looks for:
          - All submit runs with DBR >=11.3 and data_security_mode:None

        It also combines several submit runs under a single pseudo_id based on hash of the submit run configuration.
        Subsequently, a list of all the incompatible runs with failures are stored in the
        `$inventory.submit_runs` table."""
        ctx.submit_runs_crawler.snapshot()

    @job_task
    def crawl_cluster_policies(self, ctx: RuntimeContext):
        """This module scans through all the Cluster Policies and get the necessary information

        It looks for:
          - Clusters Policies with Databricks Runtime (DBR) version earlier than 11.3

          Subsequently, a list of all the policies with matching configurations are stored in the
        `$inventory.policies` table."""
        ctx.policies_crawler.snapshot()

    @job_task(cloud="azure")
    def assess_azure_service_principals(self, ctx: RuntimeContext):
        """This module scans through all the clusters configurations, cluster policies, job cluster configurations,
        Pipeline configurations, Warehouse configuration and identifies all the Azure Service Principals who has been
        given access to the Azure storage accounts via spark configurations referred in those entities.

        It looks in:
          - all those entities and prepares a list of Azure Service Principal embedded in their configurations

        Subsequently, the list of all the Azure Service Principals referred in those configurations are saved
        in the `$inventory.azure_service_principals` table."""
        if ctx.is_azure:
            ctx.azure_service_principal_crawler.snapshot()

    @job_task
    def assess_global_init_scripts(self, ctx: RuntimeContext):
        """This module scans through all the global init scripts and identifies if there is an Azure Service Principal
        who has been given access to the Azure storage accounts via spark configurations referred in those scripts.

        It looks in:
          - the list of all the global init scripts are saved in the `$inventory.global_init_scripts` table."""
        ctx.global_init_scripts_crawler.snapshot()

    @job_task
    def workspace_listing(self, ctx: RuntimeContext):
        """Scans the workspace for workspace objects. It recursively list all sub directories
        and compiles a list of directories, notebooks, files, repos and libraries in the workspace.

        It uses multi-threading to parallelize the listing process to speed up execution on big workspaces.
        It accepts starting path as the parameter defaulted to the root path '/'."""
        ctx.workspace_listing.snapshot()

    @job_task(depends_on=[crawl_grants, workspace_listing])
    def crawl_permissions(self, ctx: RuntimeContext):
        """Scans the workspace-local groups and all their permissions. The list is stored in the `$inventory.permissions`
        Delta table.

        This is the first step for the _group migration_ process, which is continued in the `migrate-groups` workflow.
        This step includes preparing Legacy Table ACLs for local group migration."""
        permission_manager = ctx.permission_manager
        permission_manager.reset()
        permission_manager.snapshot()

    @job_task
    def crawl_groups(self, ctx: RuntimeContext):
        """Scans all groups for the local group migration scope"""
        ctx.group_manager.snapshot()


class Failing(Workflow):
    def __init__(self):
        super().__init__('failing')

    @job_task
    def failing_task(self, ctx: RuntimeContext):
        """This task always fails. It is used to test the failure handling of the framework."""
        attempt = ctx.named_parameters.get("attempt", "0")
        if int(attempt) > 0:
            logger.info("This task is no longer failing.")
            return
        logger.warning("This is a test warning message.")
        logger.error("This is a test error message.")
        raise ValueError("This task is supposed to fail.")
